import UIKit
import Combine


public protocol Searchable {
  static var scopeName: String { get }
}


public protocol SearchQueryParam: Equatable {
  associatedtype FilterParams
}

public protocol SearchableData: Searchable {
  associatedtype QueryParam: SearchQueryParam
  
  static func validQueries(forSearchString searchString: String) -> [QueryParam]
  static func ~=(lhs: Self, rhs: Self.QueryParam) -> Bool
}


enum SearchDataType<Data: SearchableData> {
  case history([Data.QueryParam])
  case queries([Data.QueryParam])
  case data([Data])
  
  var count: Int {
    switch self {
    case .queries(let queries), .history(let queries):
      return queries.count
    case .data(let data):
      return data.count
    }
  }
}


protocol SearchQueryDescribable {
  var title: String { get }
  var detail: String { get }
  
  var queryDescription: String { get }
  
  func attributedString(with font: UIFont) -> NSAttributedString
}

protocol SearchQueryPersistable {
  var persistantData: [String: Any] { get }
  init?(persistantData: [String: Any])
}


public let SearchQueryPesistableQueryKey = "SearchQueryPesistableQuery"


protocol SearchResultListViewController: class {
  associatedtype Data: SearchableData
  
  var currentData: SearchDataType<Data> { get set }
  
  var historyItems: [Data.QueryParam] { get set }
  
  var queryDidSelectHandler: ((Data.QueryParam) -> Void)? { get set }
}


class AnySearchResultListViewController<Data: SearchableData>: SearchResultListViewController {
  private let dataGetter: () -> SearchDataType<Data>
  private let dataSetter: (SearchDataType<Data>) -> Void
  
  private let handlerGetter: () -> (((Data.QueryParam) -> Void)?)
  private let handlerSetter: (((Data.QueryParam) -> Void)?) -> Void
  
  private let historyItemsGetter: () -> [Data.QueryParam]
  private let historyItemsSetter: ([Data.QueryParam]) -> Void
  
  var currentData: SearchDataType<Data> {
    get {
      return dataGetter()
    }
    set {
      dataSetter(newValue)
    }
  }
  
  var queryDidSelectHandler: ((Data.QueryParam) -> Void)? {
    get {
      return handlerGetter()
    }
    set {
      handlerSetter(newValue)
    }
  }
  
  var historyItems: [Data.QueryParam] {
    get {
      return historyItemsGetter()
    }
    set {
      historyItemsSetter(newValue)
    }
  }
  
  init<P: SearchResultListViewController>(_ searchResultListViewController: P) where P.Data == Data {
    self.dataSetter = { [weak searchResultListViewController] newValue in
      searchResultListViewController?.currentData = newValue
    }
    self.dataGetter = { [weak searchResultListViewController] in
      return searchResultListViewController?.currentData ?? SearchDataType.data([])
    }
    
    self.handlerGetter = { [weak searchResultListViewController] in
      return searchResultListViewController?.queryDidSelectHandler ?? { _ in }
    }
    self.handlerSetter = { [weak searchResultListViewController] newValue in
      searchResultListViewController?.queryDidSelectHandler = newValue
    }
    
    self.historyItemsGetter = { [weak searchResultListViewController] in
      return searchResultListViewController?.historyItems ?? []
    }
    self.historyItemsSetter = { [weak searchResultListViewController] newValue in
      searchResultListViewController?.historyItems = newValue
    }
  }
}


class SearchDataViewController<Data: SearchableData>: UIViewController where Data.QueryParam: SearchQueryDescribable, Data.QueryParam: SearchQueryPersistable {
  //    @Published var list: SearchDataType<Data>
  let list: CurrentValueSubject<SearchDataType<Data>, Never>
  let resultViewController: AnySearchResultListViewController<Data>
  @Published var currentQuery: Data.QueryParam? = nil
//    let currentQuery: CurrentValueSubject<Data.QueryParam?, Never> = CurrentValueSubject(nil)
  
  let searchResultsViewController: UIViewController
  
  
  init<ResultListController: UIViewController>(resultViewController: ResultListController)
    where ResultListController: SearchResultListViewController, ResultListController.Data == Data {
      self.resultViewController = AnySearchResultListViewController(resultViewController)
      self.searchResultsViewController = resultViewController
      self.list = CurrentValueSubject(.history(resultViewController.historyItems.reversed()))
      super.init(nibName: nil, bundle: nil)
  }
  
  required init?(coder aDecoder: NSCoder) {
    fatalError()
  }
  
  
  @objc func searchFieldTextDidChanged(_ searchField: UITextField) {
    currentQuery = nil
    
    let data: SearchDataType<Data>
    defer {
      list.send(data)
    }
    
    guard let searchQuery = searchField.text, !searchQuery.isEmpty else {
      data = .history(resultViewController.historyItems.reversed())
      return
    }
    
    let validQueries = Data.validQueries(forSearchString: searchQuery)
    data = .queries(validQueries)
  }
}





